#include <ctype.h>                                      /* char types */
#include <errno.h>					/* errors */
#include <signal.h>					/* unix signals */
#include <stdio.h>                                      /* general includes */
#include <stdlib.h>                                     /* standard library */
#include <string.h>                                     /* string processing */
#include <sys/types.h>                                  /* types */
#include <unistd.h>					/* unix std routines */

#ifdef HAS_NCURSES
    #include <ncurses.h>
#else
    #include <curses.h> 
#endif

#if !defined(TRUE)
    #define TRUE        1
#endif
#if !defined(FALSE)
    #define FALSE       0
#endif

/* macros */
/*#define currentLoc(line, col) ((line) * BASE +((col)/3)) */
#define MAXY() ((LINES) - 3)			/* macro for maxy lines       */
#define CTRL_AND(c) ((c) & 037)			/* macro to use the ctrl key  */
#define cursorLoc(line,col,editHex) (((line)*BASE) + ((col)/((editHex)?3:1)))
#define llalloc() (struct linkedlist *) malloc(sizeof(struct linkedlist))

/* constants */
#define VERSION		"1.21"			/* the version of the source  */
#define BASE    	16			/* the base for the number    */
#define MIN_COLS        78                      /* screen has to be 78< cols  */
#define MIN_LINES       10                      /* there need to be 10< lines */
#define KEY_TAB 	9			/* value for the tab key      */

typedef struct                                  /* struct that holds windows  */
{   WINDOW  *hex,
            *ascii,
            *address,
            *scrollbar,
            *hex_outline,
            *ascii_outline;
}WINS;

struct linkedlist				/* linked list structure      */
{   int loc;
    int val;
    struct linkedlist *next;
};

struct Stack					/* struct to be used for stack*/
{   int    savedVal, 
	   currentLoc;
    struct Stack *prev;
    struct linkedlist *llist;
};

FILE *fpIN, *fpOUT;				/* global file ptrs */

extern FILE *fpIN, *fpOUT;                      /* global file ptrs */
extern int MAXY;				/* external globals           */
extern int CUR_COL;
extern int CUR_LINE;
typedef struct linkedlist hexList;		/* alias name to hexList      */
typedef struct Stack      hexStack;		/* alias name to hexStack     */

/* function prototypes */

/* file.c */
void outline(FILE *fp, WINS *win, hexList *head, int linenum, bool printHex);
int maxLoc(FILE *fp);
void print_usage();
int maxLines(int len);
int openfile(WINS *win, char *fpINfilename);
void savefile(WINS *win,hexList *head, char *fpINfilename, char *fpOUTfilename);
int hexSearch(FILE *fp, int ch[], int startfp, int length, hexList *head);
int gotoLine(WINS *win, FILE *fp, int currLoc, int gotoLoc, int maxlines, hexList *head, int printHex, int editHex, WINDOW *windows);

/* hexcurse.c */
int wacceptch(WINS *windows, hexList *head, int len, char *fpINfilename, char *fpOUTfilename, bool printHex);
void winscroll(WINS *win, WINDOW *, hexList *head, int, int, bool);
int parseArgs(WINS *windows, hexList *head, int, char *argv[], char *fpINfilename, char *fpOUTfilename, bool printHex);
void printDebug(hexList *head, int loc);

/* llist.c */
hexList *deleteNode(hexList *head, int loc);
hexList *insertItem(hexList *head, int loc, int val);
int searchList(hexList *head, int loc);
void writeChanges(WINS *win, FILE *fpIN, FILE *fpOUT, hexList *head, char *fpINfilename, char *fpOUTfilename);
hexList *freeList(hexList *head);

/* screen.c */
void init_menu(WINS *windows);
void exit_err(char *err_str);
void init_screen(void);
void screen_exit(int exit_val);
void init_fkeys();
void checkScreenSize(WINS *win);
void refreshall(WINS *win);
WINDOW *drawbox(int y, int x, int height, int width);
void scrollbar(WINS *windows, float currentLine, float maxLines);
void printHelp(WINS *win);
void winscroll(WINS *win, WINDOW *, hexList *head, int, int, bool);

/* stack.c */
void createStack(hexStack *stack);
void pushStack(hexStack **stack, hexStack *tmpStack);
void popStack(hexStack **stack);
void smashDaStack(hexStack **stack);
