/******************************************************************************\
 *  Copyright (C) 2001 writen by Jewfish and Armoth 
 *  Description: this codes allows a user to view and edit the hexadecimal and*
 *		 and ascii values of a file.  The curses library is used to   *
 *		 display and manipulate the output.  See the README file      *
 *		 included for more information.				      *
 *									      *
 *  This program is free software; you can redistribute it and/or modify      *
 *  it under the terms of the GNU General Public License as published by      *
 *  the Free Software Foundation; either version 2 of the License, or	      *
 *  (at your option) any later version.					      *
 *									      *
 *  This program is distributed in the hope that it will be useful,	      *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of	      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	      *
 *  GNU General Public License for more details.			      *
 *									      *
 *  You should have received a copy of the GNU General Public License	      *
 *  along with this program; if not, write to the Free Software		      *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *									      *
\******************************************************************************/

#include "hex.h"					/* custom function
							   header */
#define DEBUG_LLIST
//#define DEBUG_GOTO

#define KEY_PGDN 338
#define KEY_PGUP 339

int CUR_LINE, CUR_COL;					/* env settings */

int main(int argc, char *argv[])			/* main program */
{
    bool printHex;					/* address format */
    WINS *windows;					/* window structure */
    hexList *head;					/* linked list struct */
    int  row, col, len, x, retval = 1, val;		/* counters, etc.*/
    
    char *fpINfilename = NULL, 				/* in and out file */
	 *fpOUTfilename = NULL;				/* name pointers   */
 
    head = llalloc();					/* malloc list space */
    windows = (WINS *) malloc(sizeof(WINS));		/* malloc windows */
    fpINfilename = (char *) malloc(81);			/* allocate in and */
    fpOUTfilename = (char *) malloc(81);		/* out file name ptrs */
    printHex = TRUE;
							/* get cmd line args */
    len = parseArgs(windows, head, argc, argv, fpINfilename, fpOUTfilename,
		    printHex);
     
    slk_init(0);					/* init menu bar */
    init_screen();					/* init visuals */
    init_fkeys();					/* define menu bar */
    init_menu(windows);					/* init windows    */ 

    while(retval)
    {
        head = freeList(head);				/* free & init head */
							/* print origin loc */
	mvwprintw(windows->hex_outline, 0, 1, "00000000");
    
	if (fpIN != NULL)				/* if no infile... */	
	{
            len = maxLoc(fpIN);				/* get last file loc */
	    val = maxLines(len); 			/* max file lines */
            for(x = 0; x <= MAXY() && x <= val; x++)	/* output lines */
		outline(fpIN, windows, head, x, printHex); 
	}

	wmove(windows->hex, 0, 0);			/* cursor to origin */
	row = col = 0;					/* init row & col */
    
	refreshall(windows);				/* refresh all wins */
	doupdate();					/* update screen */

	mvwaddch(windows->scrollbar, 1, 0, ACS_CKBOARD);/* clear scroller */
							/* get user input */
	retval = wacceptch(windows, head, len, fpINfilename, fpOUTfilename,
			   printHex); 
    }
    
    screen_exit(0);					/* end visualizations */
    return 0;						/* return */
}

/********************************************************\
 * Description: accepts a character from the user and	*
 *		interprets it.				*
 * Returns:	retval (if quit program or no)		*
\********************************************************/
int wacceptch(WINS *win, hexList *head, int len, char *fpINfilename, 
	      char *fpOUTfilename, bool printHex)
{
    int row, col, count, maxlines, val, tmpval, cl,	/* counters, etc. */
        ch[17],						/* holds search string*/
	eol = 47,					/* end of line pos */
	gotoLoc =0,					/* goto location    */
	prevScroll = 0,					/* previous scrlbar */
	currentLine = 0,				/* current line */
	tmp;
    char temp[80],
	 SearchStr[13];

    short int key;					/* key capture */
    WINDOW *windows;					/* window pointer */
    hexStack *stack;					/* used for stack */
    hexStack *tmpStack;					/* used for stack */
    bool editHex = TRUE;				/* val for editing */
    maxlines = maxLines(len);				/* lines in file */
    windows = win->hex;					/* curr edit window */
							/* get keys til exit */

    /*createStack(stack);*/				/* init the stack */
    stack = NULL;
    bzero(temp, 80);					/* zero out string */
    SearchStr[0] = NULL;					

    while(((key = wgetch(windows)) != KEY_F(8)) && (key != CTRL_AND('x')))
    {
	getyx(windows, row, col);			/* curent cursor loc */
        CUR_LINE = LINES;				/* curr term params */
	CUR_COL = COLS;
	/*checkScreenSize(win);	*/			/* check term size */

	switch(key) {					/* check keypress */

	case KEY_UP:					/* if UP... */
		if (currentLine > 0) 			/* move up... */
		{   
		    currentLine--;
		    wmove(windows, --row, col);
		}

		if((currentLine >= 0) && (row < 0))	/* scroll up... */
		    winscroll(win, windows, head, -1, currentLine, printHex);
		break;

	case KEY_DOWN:					/* if DOWN... */
		if (cursorLoc(currentLine + 1, col, editHex) < len) 
		{   
		    if (currentLine < maxlines)		/* move down... */
		    {
		        wmove(windows, ++row, col);
		        currentLine++;
		    }
							/* scroll down... */
		    if((row > MAXY()) && (currentLine <= maxlines))
		        winscroll(win, windows, head, 1, currentLine, printHex);
		}
		break;

	case KEY_BACKSPACE:
	case KEY_LEFT:					/* if LEFT or BACK... */
		if ((col == 0) && (currentLine != 0))   /* move up... */
		{   
	  	    currentLine--;
		    if(row == 0)			/* scroll up... */
		    {
			winscroll(win, windows, head,-1, currentLine, printHex);
			row++;
		    }
		    wmove(windows, --row, eol-1);
		}

		wmove(windows, row, --col);		/* move left... */
		if (editHex)
		{
		    if((col + 1) % 3 == 0)
		        wmove(windows, row, --col);
		}
		break;

	default:					/* if other key... */
							/* if key we want... */
		if(isprint(key) && ((editHex && isxdigit(key)) || !editHex))
		{
		    if((cl=cursorLoc(currentLine, col, editHex))< len) 
		    {   
		
							/* if not in ll... */
			if((val = searchList(head, cl)) == -1)
		  	{
			    fseek(fpIN, cl, SEEK_SET);	/* get val from file */
			    val = fgetc(fpIN);
			}

			wattron(win->hex, A_BOLD);
			wattron(win->ascii, A_BOLD);
							/* output it */
			wprintw(windows, "%c", editHex ? toupper(key): key);

			tmpval = val;			/* val b4 key press */

			if(editHex)			/* if in hex win... */
			{
			    if(key >= 65 && key <= 70)	/* get correct val */
			    	key -= 7;
			    else if(key >= 97 && key <= 102)
			    	key -= 39;
			    key -= 48;
			
			    if ((col % 3) == 0)		/* compute byte val */
                            	val = (key * BASE) + (val % BASE);
			    else if ((col % 3) == 1)
			    	val = (val - ((val + BASE) % BASE) + key);
			}
			else				/* else... */
			    val = key;			/* val is key pressed */

			if (editHex)			/* update ascii win */
			{
			    wmove(win->ascii, row, (col/3));	
			    wprintw(win->ascii, "%c", isprint(val) ? val : 46);
			    wmove(win->hex, row, col);
			    wrefresh(win->ascii);
			}
			else				/* update hex win */
			{
			    wmove(win->hex, row, (col*3));	
			    wprintw(win->hex, "%02X", val);
			    wmove(win->ascii, row, col);
			    wrefresh(win->hex);
			}

			wattrset(win->hex, A_NORMAL);
			wattrset(win->ascii, A_NORMAL);
			
							/* edit list */
			head = insertItem(head, cl, val);
		/* calloc() is used because it NULLS out all returned memory */
    			tmpStack = (hexStack *) calloc(1, sizeof(hexStack));
			tmpStack->currentLoc = cl;
			tmpStack->llist	     = head;
			tmpStack->savedVal   = tmpval;
			tmpStack->prev       = NULL;
			pushStack(&stack, tmpStack);
		    }					/* continue to next
							    case */

	case KEY_RIGHT:					/* if RIGHT... */
		    if (cursorLoc(currentLine, col, editHex) < len) 
		    {   
			wmove(windows, row, ++col);	/* move right */
		        if (cursorLoc(currentLine, col,editHex) == len)
			    wmove(windows, row, --col);
							/* move down */
		        if((col == eol) && (currentLine < maxlines))
		        {
		            currentLine++;
		            if(row < MAXY())
	  	                wmove(windows, ++row, 0);
	                    else			/* scroll down */
	                    {
			        winscroll(win, windows, head, 1, currentLine,
				          printHex);
	 		        wmove(windows, row, 0);
	                    }
		        }
		        if (editHex)			/* adjust for hex win */
		            if((col + 1) % 3 == 0)
		                wmove(windows, row, ++col);
		    }

							/* if end of file... */
							/* adjust properly   */
		    if (cursorLoc(currentLine, col, editHex) == len)
			wmove(windows, row, (col = col - 2));
		}
		break;

	case CTRL_AND('u'):
	case KEY_PGUP:					/* if KEY_PGUP... */
		if (currentLine == row) {		/* if first page */
		    currentLine = 0;			/* just move to top */
		    wmove(windows,0,0);
		    break;
		}
		else
		    currentLine -= (2*MAXY());

	case CTRL_AND('d'):
	case KEY_PGDN:					/* if KEY_PGDN... */
		getyx(windows, row, col);		/* current location */

							/* if EOF < page away */
		if((maxlines - currentLine) <= MAXY()) {
							/* inc line til end */
		    while(cursorLoc(currentLine + 1, col, editHex) < len) {
		        if (currentLine < maxlines){/* move down... */
		            row++;
		            currentLine++;
		        }
							/* scroll down... */
	                if((row > MAXY()) && (currentLine <= maxlines)) {
	                    winscroll(win, windows, head, 1, currentLine, printHex);
			    row = MAXY();		/* last row/col */
		            col = (editHex) ? 3 * ((len % BASE) - 1):(len % BASE) - 1;
			}
		    }
							/* move to EOF */
		    if((len - cursorLoc(currentLine, col, editHex) <= BASE) &&
		       (currentLine < maxlines) && row >= MAXY()) {
		        winscroll(win, windows, head, 1, ++currentLine, printHex);
		    }

		    row += (maxlines - currentLine);	/* adj to last line */
		    currentLine = maxlines;
							/* last column */
		    col = (editHex) ? 3 * ((len % BASE) - 1):(len % BASE) - 1;

		    wmove(windows,row,col);		/* move cursor */
 		    break;
		}
		
		currentLine += MAXY();

		if(currentLine > maxlines)		/* adjust currentLine */
		    currentLine = maxlines;
		else if(currentLine < 0)
		    currentLine = 0;


		val = currentLine - row;		/* get val */
		if(val < 0)
		   row = val = 0;

		wclear(win->hex);			/* clear windows */
		wclear(win->ascii);
		wclear(win->address);

		wmove(win->hex, 0, 0);			/* position cursors */
		wmove(win->ascii, 0, 0);
		wmove(win->address, 0, 0);
							/* output lines */
		for(count = 0; count <= MAXY() && (count + val) <= maxlines;
		    count++)
		    outline(fpIN, win, head, (count + val), printHex);
		
		wmove(windows, row, col);		/* restore cursor */
		/*refreshall(win);*/ 			/* refresh */
		wnoutrefresh(win->scrollbar);
    		wnoutrefresh(win->ascii);
    		wnoutrefresh(win->address);
    		wnoutrefresh(win->hex);
		doupdate();				/* update screen */
	
		break;
	case CTRL_AND('o'):
	case KEY_F(3):					/* if F3 or ^o... */
		if (openfile(win, fpINfilename))	/* open file */
	 	    return TRUE;			/* TRUE if worked */
		break;

	case CTRL_AND('s'):				/* if F2 or ^s... */
	case KEY_F(2):					/* save the file */	
		savefile(win, head, fpINfilename, fpOUTfilename);
		break;

	case CTRL_AND('f'):
	case KEY_F(5):
 		/* SeachStr stores the last searched string into the format *\
		\* "(XXXXXXX...)" with 10 being the max chars shown   */
		if (temp[0] != NULL)
		{
		    bzero(SearchStr, 13);
		    strcat(SearchStr, "(");
		    if (strlen(temp) <= 10)
			strncat(SearchStr, temp, strlen(temp));
		    else
		    {
			strncat(SearchStr, temp, 7);
			strcat(SearchStr, "...");
		    }
		    strcat(SearchStr, ")");
		}

		wmove(win->hex_outline, LINES-1, 19);   /* output prompt    */
    		wclrtoeol(win->hex_outline);
    		mvwprintw(win->hex_outline, LINES - 1, 1, 
		"Enter %s value %s: ", 
		(editHex) ? "hex" : "ascii", SearchStr);

		echo();					/* echo chars       */
		wscanw(win->hex_outline, "%s", temp);
		noecho();

		val = 0;
							/* parse out input */
	        for (count = 0; temp[count] != 0 && count < 80; count++)
		    if(!editHex)
			ch[count] = temp[count];
		    else
							/* if hex digit */
	 		if (isxdigit(key = temp[count])) {
			    if(key >= 65 && key <= 70)	/* get correct val */
			    	key -= 7;
			    else if(key >= 97 && key <= 102)
			    	key -= 39;
			    key -= 48;
			
			    if ((count % 2) == 0)	/* compute byte val */
                            	tmp = (key * BASE);
			    else 
			    	ch[(count - 1) / 2] = tmp + key;
			}
			else				/* if not hex, bad! */
			    val = -1;

		if((count % 2 > 0) && (editHex))	/* add last byte on */
			    ch[(count + 1) / 2] = tmp;

		if(val != -1)				/* if val checks out */
							/* search for it */
		    val = hexSearch(fpIN, ch, cursorLoc(currentLine, col,
			  editHex) + 1, (editHex) ? ((count + 1)/2):count,head);

		if (val == -1) {			/* if nothing came up */
		    for (count = 1; count < 59; count++)/* output message */
                        mvwaddch(win->hex_outline, LINES-1, count, 32);
        	    mvwaddstr(win->hex_outline, LINES-1,1,"Value not found!");

        	    wrefresh(win->hex_outline);         /* refresh window */

        	    sleep(2);                           /* wait 2 secs */

        	    for (count = 1; count < 58; count++)/* restore border */
            	        mvwaddch(win->hex_outline, LINES-1, count, ACS_HLINE);
        	    mvwaddch(win->hex_outline, LINES-1, 58, ACS_LRCORNER);

		    wrefresh(win->hex_outline);
		}
		else {
                    getyx(windows, row, col);
							/* goto found loc */
                    currentLine = gotoLine(win, fpIN,
                                           cursorLoc(currentLine, col, editHex),
                                           val, maxlines, head, printHex,
                                           editHex, windows);

		}
		break;


	case CTRL_AND('a'):
	case KEY_F(6):					/* if F6, ^a, ^d... */
		printHex = (!printHex);			/* reverse printHex */

		getyx(windows, row, col);		/* current location */
		row = currentLine - row;		/* compute top line */

		wmove(win->address, 0, 0);		/* move to origin */
							/* write out values */
		for(count = 0; count <= MAXY() && count <= maxlines ; count++)
		    wprintw(win->address, (printHex) ? "%08X ":"%08d ",
			   (count + row) * BASE);

							/* update menu button */
		slk_set(6, (printHex) ? "HexAdres":"DecAdres", 1);
		slk_noutrefresh();			/* refresh */
		wnoutrefresh(win->address);
		break;

 	case CTRL_AND('g'):
	case KEY_F(4):					/* if F4 or ^g... */
		wmove(win->hex_outline, LINES-1, 21);   /* output prompt    */
    		wclrtoeol(win->hex_outline);
    		mvwprintw(win->hex_outline, LINES - 1, 1, 
		"Enter %s location: ", (printHex) ? "HEX":"decimal");

		echo();					/* echo chars       */
		wscanw(win->hex_outline, (printHex) ? "%X":"%d",&gotoLoc);
		noecho();				/* disable echoing  */

		if ((gotoLoc < 0) || (gotoLoc > len-1))
		{   
		    for (count = 1; count < 59; count++)/* output message */
                        mvwaddch(win->hex_outline, LINES-1, count, 32);
        	    mvwaddstr(win->hex_outline, LINES-1,1,"Invalid location!");

        	    wrefresh(win->hex_outline);         /* refresh window */

        	    sleep(2);                           /* wait 2 secs */

        	    for (count = 1; count < 58; count++)/* restore border */
            	        mvwaddch(win->hex_outline, LINES-1, count, ACS_HLINE);
        	    mvwaddch(win->hex_outline, LINES-1, 58, ACS_LRCORNER);

		    wrefresh(win->hex_outline);
		}
		else {
                    getyx(windows, row, col);
							/* goto found loc */
                    currentLine = gotoLine(win, fpIN,
                                           cursorLoc(currentLine, col, editHex),
                                           gotoLoc, maxlines, head, printHex,
                                           editHex, windows);
	 	}
		break;

	case KEY_TAB:
	case KEY_F(7):					/* if F7, TAB, ^i... */
		if (editHex)				/* already in hex win */
		{
		    windows = win->ascii;		/* change to ascii*/
		    eol = 16;
	 	    wmove(windows, row, (col/3));
		    slk_set(7, "Asc Edit", 1);
		}
		else					/* already in ascii */
		{
		    windows = win->hex;			/* change to hex win */
		    eol = 47;
	 	    wmove(windows, row, (col*3));	
		    slk_set(7, "Hex Edit", 1);
		}

		editHex = (!editHex);			/* change test val */
		slk_noutrefresh();			/* refresh menu */
		break;

	case CTRL_AND('?'):
	case CTRL_AND('h'):
	case KEY_F(1):					/* if F1, ^?, ^h... */
		getyx(windows, row, col);		/* current location */
		wclear(win->hex);
		wclear(win->ascii);
		wclear(win->address);
		wnoutrefresh(win->hex);
    		wnoutrefresh(win->ascii);
    		wnoutrefresh(win->address);
		doupdate();				/* update screen */
		printHelp(win);				/* display the help */
		wclear(win->hex);
		wclear(win->ascii);
		wclear(win->address);
		wnoutrefresh(win->hex);
    		wnoutrefresh(win->ascii);
    		wnoutrefresh(win->address);
		doupdate();				/* update screen */
		wgetch(win->hex);			/* ne key to continue */
		wmove(win->hex, 0, 0);			/* position cursors */
		wmove(win->ascii, 0, 0);
		wmove(win->address, 0, 0);
		for(count = 0; count <= MAXY() && (count + currentLine - row) 
		    <= maxlines; count++)
		    outline(fpIN, win, head, (count+currentLine-row), printHex);
		wmove(win->hex, 0, 0);			/* position cursors */
		wmove(win->ascii, 0, 0);
		wmove(win->address, 0, 0);
    		wnoutrefresh(win->hex);
    		wnoutrefresh(win->ascii);
    		wnoutrefresh(win->address);
		wnoutrefresh(win->scrollbar);
		doupdate();				/* update screen */
	 	wmove(windows, row, col);	
		break;

	case CTRL_AND('z'):				/* ^z undo last mod */
		getyx(windows, row, col);

							/* set previous loc */
		cl = (stack == NULL) ? cl : stack->currentLoc;
		if (stack != NULL)
		{

		    /*if (stack != NULL) val = stack->savedVal;*/
		    val = stack->savedVal;
		    popStack(&stack);
		    head = deleteNode(head,cl);


                    currentLine = gotoLine(win, fpIN, 
			  	    	cursorLoc(currentLine, col, editHex),
                                    	cl, maxlines, head, printHex,
                                    	editHex, windows);
		/*
                    if (stack != NULL)
		    { 
		        mvwprintw(win->hex_outline,0,20,"                    ");
		        mvwprintw(win->hex_outline,0,10,
			      "value:%02X    location:%d  ",
		    	      stack->currentLoc, 
			      currentLine);
		    }
		*/
		
		    getyx(windows, row, col);

   		    if((searchList(head, cl)) != -1)
		    {	
			wattron(win->hex, A_BOLD);
			wattron(win->ascii, A_BOLD);
		    }
                    if (editHex)
		    {
                        wmove(win->ascii, row, (col/3));    
                        wprintw(win->ascii, "%c", isprint(val) ? val : 46);
                        wmove(win->hex, row, col);
                        wrefresh(win->ascii);

                        wprintw(win->hex, "%02X", val);
                        wmove(win->hex, row, col);      
                        wrefresh(win->hex);
		    }
		    else
		    {
                        wmove(win->hex, row, (col*3));      
                        wprintw(win->hex, "%02X", val);
                        wmove(win->ascii, row, col);
                        wrefresh(win->hex);

                        wprintw(win->ascii, "%c", isprint(val) ? val : 46);
                        wmove(win->ascii, row, col);    
                        wrefresh(win->ascii);
		    }
		    wattrset(win->hex, A_NORMAL);
		    wattrset(win->ascii, A_NORMAL);


		}
		break;

#ifdef DEBUG_LLIST
	case CTRL_AND('q'):
	    printDebug(head, -1);
	    break;

	}
#endif

	getyx(windows, row, col);			/* get cur row/col */
    
							/* print cur location */
	mvwprintw(win->hex_outline, 0, 1, (printHex) ? "%08X":"%08d",
	   	  cursorLoc(currentLine, col, editHex));
	
	wnoutrefresh(win->hex_outline);			/* refresh outline */
	
							/* calc scroll bar */
	count = ((float)currentLine / ((float)maxlines / (float)MAXY())) + 1;
	if(count != prevScroll)
 	{
	    scrollbar(win, currentLine, maxlines);	/* change scroll bar */
	    prevScroll = count;
	}
	
	doupdate();					/* update visual */
	wmove(windows, row, col);			/* restore cursor */
    }

    return(0);						/* return */
}

void printDebug(hexList *head, int loc)
{
    FILE *tmpofp;
    hexList *tmpHead = head;

    tmpofp = fopen("debug_llist", "a+");
    tmpHead = head;

    fprintf(tmpofp, "location undone: %08X\n", loc);
    while(tmpHead != NULL) {
	fprintf(tmpofp, "head->loc: %08X   head->val: %02X (%c)\n", tmpHead->loc, tmpHead->val, tmpHead->val);

	tmpHead = tmpHead->next;
    }
    fprintf(tmpofp, "\n");

    fclose(tmpofp);
}

/********************************************************\
 * Description: parses command line arguments and 	*
 *		processes them.				*
 * Returns:	length of file				*
\********************************************************/
int parseArgs(WINS *win, hexList *head, int argc, char *argv[], 
	      char *fpINfilename, char *fpOUTfilename, bool printHex)
{
    extern char *optarg;				/* extern vars for */
    extern int optind, opterr, optopt;			/* getopt()	   */
 
    int val;						/* counters, etc. */

    opterr = FALSE;					/* no getopt() stderr */

							/* get args */
    while ((val = getopt(argc, argv, "i:o:")) != -1) 
    {
	switch(val)					/* test args */
        {
							/* infile */
	    case 'i':	strncpy(fpINfilename, optarg, 80);
			if ((fpIN = fopen(optarg, "r")) == NULL)
			    exit_err("Could not open file");
			break;
							/* outfile */
	    case 'o':   strncpy(fpOUTfilename, optarg,80);
			break;
							/* help/invalid args */ 
	    case '?':	if ((optopt == 'h') || (optopt == '?'))
			{
			    print_usage();		/* output help */
			    exit(0);			/* exit */
			}
			else				/* illegal option */
			{
			    printf("\nhexcurse: Invalid option\n");
			    print_usage();
			    exit(-1);
			}
        }
    }
 
    return((fpIN != NULL) ? maxLoc(fpIN):0);		/* return file length */
}
