#include "hex.h"

/******************************************************\
 * Description: Inits the menu and the user interface *
 *		that the hex editor uses.             *
\******************************************************/
void init_menu(WINS *windows)
{   
    int x;
    WINDOW *win;

    for(x = 1; x < 24; x++)				/* blocks all signals */
	signal(x, SIG_IGN);				/* ignores the sigs   */

    /*signal(SIGCHLD,  SIG_DFL);*/
    signal(SIGSEGV,  SIG_DFL);				/* reset segfault sig */

    windows->hex_outline = drawbox(0, 0, LINES, 59);	/* draw the windows   */
    windows->ascii_outline = drawbox(0, 60, LINES, 18);
    windows->scrollbar = newwin(LINES, 1, 0, 59);

    windows->hex = newwin(LINES - 2, 48, 1, 10);
    windows->ascii = newwin(LINES - 2, 16, 1, 61);
    windows->address = newwin(LINES - 2, 9, 1, 1);

    waddch(windows->scrollbar, ACS_UARROW);
    wattron(windows->scrollbar, A_REVERSE);
    waddch(windows->scrollbar, ACS_DIAMOND);
    wattroff(windows->scrollbar, A_REVERSE);
    for(x = 0; x < LINES - 2; x++)
        waddch(windows->scrollbar, ACS_CKBOARD);
    waddch(windows->scrollbar, ACS_DARROW);
    wrefresh(windows->scrollbar);

    cbreak();
    noecho();						/* disable echoing    */
    raw();						/* enable raw kb input*/
    
    win = windows->hex;
    for(x = 0; x <= 2; x++)				
    {
	idlok(win, TRUE);				/* enable scrolling   */

        keypad(win, TRUE);				/* enable the keypad  */

	win = (x == 0) ? windows->ascii:windows->address;
    }
    CUR_LINE = LINES;					/*store cur term lines*/
    CUR_COL = COLS;					/*store cur term cols */

}

/******************************************************\
 * Description:	exit the program printing out an error*
 *		from the errno value.	 	      *
\******************************************************/
void exit_err(char *err_str)
{   endwin(); 						/* end curses screen  */
    perror(err_str);					/* print errno        */
    printf("\n\n");
    exit(EXIT_FAILURE);					/* return with exitval*/
}

/******************************************************\
 * Description: Initialize the curses graphical screen*
 *		If the screen does not support curses *
 *		exit gracefully.		      *
\******************************************************/ 
void init_screen(void)
{   if ((initscr()) == NULL)				/*couldn't init screen*/
    {   perror("initscr");				/* print out error    */
        printf("exiting\n");
	exit(EXIT_FAILURE);
    }
}

/******************************************************\
 * Description:	Terminate the curses screen and exit  *
 *		out of the program.		      *
\******************************************************/
void screen_exit(int exit_val)
{   endwin();						/* end curses screeen */
    printf("\n\n");
    exit(exit_val);
}

/******************************************************\
 * Description: Initialize function key menu interface*
\******************************************************/ 
void init_fkeys()
{
    slk_set(1, "Help",		1);			/* init menu tabs */
    slk_set(2, "Save",		1);
    slk_set(3, "Open",	 	1);
    slk_set(4, "Goto",		1);
    slk_set(5, "Find", 		1);
    slk_set(6, "HexAdres",	1);
    slk_set(7, "Hex Edit",	1);
    slk_set(8, "Quit",		1);
    slk_noutrefresh();					/* refresh the screen */
}

/******************************************************\
 * Description:	Determines the screen size of the the *
 *		terminal and prompts for the user to  *
 *		change the screen.		      *
\******************************************************/
void checkScreenSize(WINS *win)
{   int y, x;						/* counter variables  */
    
    if ((CUR_COL < MIN_COLS) || (CUR_LINE < MIN_LINES)) /* check for term size*/
    {    
	werase(win->hex_outline);			/* clear the windows  */
	werase(win->ascii_outline);
	werase(win->hex);
	werase(win->ascii);
	getmaxyx(win->hex, y, x), 			/* get size of hex win*/
	mvwprintw (win->hex, y/2-15, x/2-15, "Please resize your terminal!!");
	/*
        while ((CUR_COL < MIN_COLS) || (CUR_LINE < MIN_LINES))  
	{    CUR_COL = COLS;				loop till the screen
	
	     CUR_LINE = LINES;				size is correct    
	     sleep(3);
	}
	*/
         
	screen_exit(-1);
    }
    /*else wprintw(win->hex, "%d %d", COLS, LINES);*/
}

/******************************************************\
 * Description:	refresh all of the curses windows     *
\******************************************************/
void refreshall(WINS *win)
{   slk_noutrefresh();					/* refresh fkeys menu */
    wnoutrefresh(win->hex_outline);			/* refresh all windows*/
    wnoutrefresh(win->ascii_outline);
    wnoutrefresh(win->scrollbar);
    wnoutrefresh(win->ascii);
    wnoutrefresh(win->address);
    wnoutrefresh(win->hex);
}

/******************************************************\
 * Description:	draws a box with the given parameters *
 *		these boxes are used to create the    *
 *		user interface.  The function returns *
 *		the newly created window	      *
\******************************************************/
WINDOW *drawbox(int y, int x, int height, int width)
{
    int n;

    WINDOW *win;
    win = newwin(height, width, y, x);			/* create a new window*/
    
    mvwaddch(win, 0, 0, ACS_ULCORNER);			/* add window corners */
    mvwaddch(win, height - 1, 0, ACS_LLCORNER);
    mvwaddch(win, 0, width - 1, ACS_URCORNER);
    mvwaddch(win, height - 1, width - 1, ACS_LRCORNER);
    
    for(n = 1; n < width - 1; n++)			/* construct the boxes*/
    {
	mvwaddch(win, 0, n, ACS_HLINE);
	mvwaddch(win, height - 1, n, ACS_HLINE);
    }
	
    for(n = 1; n < height - 1; n++)			 
    {
	mvwaddch(win, n, 0, ACS_VLINE);
	mvwaddch(win, n, width - 1, ACS_VLINE);
    }

    wrefresh(win);					/* refresh the screen */

    return win;						/*return created win  */
}

/******************************************************\
 * Description:	construct a scroll bar that is used   *
 *		for scrolling through the hex editor  *
\******************************************************/ 
void scrollbar(WINS *windows, float currentLine, float maxLines)
{
    float x;
							/* find row */
    x = ((float)currentLine / ((float)maxLines / ((float)MAXY()))) + 1;
    if(x <= MAXY() || x > 0)
    {
        wattron(windows->scrollbar, A_REVERSE);		/* set attribs on bar */
        mvwaddch(windows->scrollbar, (int)x, 0, ACS_DIAMOND);
        wattroff(windows->scrollbar, A_REVERSE);
        wnoutrefresh(windows->scrollbar);		/* refresh the bar    */
        mvwaddch(windows->scrollbar, (int)x, 0, ACS_CKBOARD);
    }
}

/********************************************************\
 * Description: outputs key command help                *
 * Returns:     none                                    *
\********************************************************/
void printHelp(WINS *win)
{
    wmove(win->address,0,0);
    mvwprintw(win->address, 0, 0, "Ctrl Keys");         /* print in address*/
    mvwprintw(win->address, 2, 1, "CTRL+H");
    mvwprintw(win->address, 3, 1, "CTRL+S");
    mvwprintw(win->address, 4, 1, "CTRL+O");
    mvwprintw(win->address, 5, 1, "CTRL+G");
    mvwprintw(win->address, 6, 1, "CTRL+F");
    mvwprintw(win->address, 7, 1, "CTRL+A");
    mvwprintw(win->address, 8, 1, "TAB");
    mvwprintw(win->address, 9, 1, "CTRL+X");
    mvwprintw(win->address, 11, 1, "CTRL+U");
    mvwprintw(win->address, 12, 1, "CTRL+D");
    mvwprintw(win->address, 13, 1, "CTRL+Z");

    mvwprintw(win->ascii, 0, 0, "Function Keys");       /* print in ascii  */
    mvwprintw(win->ascii, 2, 1, "Help     = F1");
    mvwprintw(win->ascii, 3, 1, "Save     = F2");
    mvwprintw(win->ascii, 4, 1, "Open     = F3");
    mvwprintw(win->ascii, 5, 1, "Goto     = F4");
    mvwprintw(win->ascii, 6, 1, "Find     = F5");
    mvwprintw(win->ascii, 7, 1, "HexAdres = F6");
    mvwprintw(win->ascii, 8, 1, "Hex Edit = F7");
    mvwprintw(win->ascii, 9, 1, "Quit     = F8");
    mvwprintw(win->ascii, 11, 1, "Page up  = PGUP");
    mvwprintw(win->ascii, 12, 1, "Page down= PGDN");

    mvwprintw(win->hex, 0, 10, "HexCurse Keyboard Commands");
    mvwprintw(win->hex, 2, 2, "Help     - you are reading it now");
    mvwprintw(win->hex, 3, 2, "Save     - saves the current file open");
    mvwprintw(win->hex, 4, 2, "Open     - opens a new file");
    mvwprintw(win->hex, 5, 2, "Goto     - goto a specified address");
    mvwprintw(win->hex, 6, 2, "Find     - search for a hex/ascii value");
    mvwprintw(win->hex, 7, 2, "HexAdres - toggle between hex/decimal address");
    mvwprintw(win->hex, 8, 2, "Hex Edit - toggle between hex/ASCII windows");
    mvwprintw(win->hex, 9, 2, "Quit     - exit out of the program");
    mvwprintw(win->hex, 11, 2, "Page up  - scrolls one screen up");
    mvwprintw(win->hex, 12, 2, "Page down- scrolls one screen down");
    mvwprintw(win->hex, 13, 2, "Undo     - reverts last modification");

    mvwprintw(win->hex, 20, 12, "Press any key to continue");
}

/********************************************************\
 * Description: scrolls window one line up or down      *
 * Returns:     none                                    *
\********************************************************/
void winscroll(WINS *win, WINDOW *windows, hexList *head, int n,
                int currentLine, bool printHex)
{
    int row, col;                                       /* row and col */

    getyx(windows, row, col);                           /* get cur row/col */

    scrollok(win->hex, TRUE);                           /* allow scrolling */
    scrollok(win->ascii, TRUE);
    scrollok(win->address, TRUE);

    wscrl(win->hex, n);                                 /* scroll each win */
    wscrl(win->ascii, n);
    wscrl(win->address, n);

    scrollok(win->hex, FALSE);                          /* disable scrolling */
    scrollok(win->ascii, FALSE);
    scrollok(win->address, FALSE);

    wmove(win->hex, (n == 1) ? MAXY():0, 0);            /* place cursor */
    wmove(win->ascii, (n == 1) ? MAXY():0, 0);
    wmove(win->address, (n == 1) ? MAXY():0, 0);

    outline(fpIN, win, head, currentLine, printHex);    /* output line */

    wnoutrefresh(win->hex);                             /* set win refreshes */
    wnoutrefresh(win->ascii);
    wnoutrefresh(win->address);

    wmove(windows, row, col);                           /* restore cursor */
}
